@@ -6,7 +6,7 @@ use experimental 'smartmatch';
 
 # ABSTRACT: Web::API - A Simple base module to implement almost every RESTful API with just a few lines of configuration
 
-our $VERSION = '1.8'; # VERSION
+our $VERSION = '1.9'; # VERSION
 
 use LWP::UserAgent;
 use HTTP::Cookies;
@@ -192,6 +192,20 @@ has 'signature_method' => (
 );
 
 
+has 'encoder' => (
+    is        => 'rw',
+    isa       => 'CodeRef',
+    predicate => 'has_encoder',
+);
+
+
+has 'decoder' => (
+    is        => 'rw',
+    isa       => 'CodeRef',
+    predicate => 'has_decoder',
+);
+
+
 has 'oauth_post_body' => (
     is      => 'rw',
     isa     => 'Bool',
@@ -255,16 +269,24 @@ sub decode {
 
     my $data;
     eval {
-        given ($content_type) {
-            when (/plain/) { $data = $content; }
-            when (/urlencoded/) {
-                foreach (split(/&/, $content)) {
-                    my ($key, $value) = split(/=/, $_);
-                    $data->{ uri_unescape($key) } = uri_unescape($value);
+        if ($self->has_decoder) {
+            $self->log('running custom decoder') if $self->debug;
+            $data = $self->decoder->($content, $content_type);
+        }
+        else {
+            given ($content_type) {
+                when (/plain/) { $data = $content; }
+                when (/urlencoded/) {
+                    foreach (split(/&/, $content)) {
+                        my ($key, $value) = split(/=/, $_);
+                        $data->{ uri_unescape($key) } = uri_unescape($value);
+                    }
+                }
+                when (/json/) { $data = $self->json->decode($content); }
+                when (/xml/) {
+                    $data = $self->xml->XMLin($content, NoAttr => 0);
                 }
             }
-            when (/json/) { $data = $self->json->decode($content); }
-            when (/xml/) { $data = $self->xml->XMLin($content, NoAttr => 0); }
         }
     };
     return { error => "couldn't decode payload using $content_type: $@\n"
@@ -280,16 +302,24 @@ sub encode {
 
     my $payload;
     eval {
-        given ($content_type) {
-            when (/plain/) { $payload = $options; }
-            when (/urlencoded/) {
-                $payload .=
-                    uri_escape($_) . '=' . uri_escape($options->{$_}) . '&'
-                    foreach (keys %$options);
-                chop($payload);
+        # custom encoder should only be run if called by Web::API otherwise we
+        # end up calling it twice
+        if ($self->has_encoder and caller(1) eq 'Web::API') {
+            $self->log('running custom encoder') if $self->debug;
+            $payload = $self->encoder->($options, $content_type);
+        }
+        else {
+            given ($content_type) {
+                when (/plain/) { $payload = $options; }
+                when (/urlencoded/) {
+                    $payload .=
+                        uri_escape($_) . '=' . uri_escape($options->{$_}) . '&'
+                        foreach (keys %$options);
+                    chop($payload);
+                }
+                when (/json/) { $payload = $self->json->encode($options); }
+                when (/xml/)  { $payload = $self->xml->XMLout($options); }
             }
-            when (/json/) { $payload = $self->json->encode($options); }
-            when (/xml/)  { $payload = $self->xml->XMLout($options); }
         }
     };
     return { error => "couldn't encode payload using $content_type: $@\n"
@@ -386,8 +416,8 @@ sub talk {
 
     if ($self->debug) {
         $self->log("uri: $method $uri");
-        $self->log("extra headers:\n" . dump(\%header)) if (%header);
-        $self->log("OAuth headers:\n" . $oauth_req->to_authorization_header)
+        $self->log("extra headers: " . dump(\%header)) if (%header);
+        $self->log("OAuth headers: " . $oauth_req->to_authorization_header)
             if ($self->auth_type eq 'oauth_header');
     }
 
@@ -460,7 +490,15 @@ sub map_options {
 
         my @missing_attrs;
         foreach my $attr (@{ $command->{mandatory} }) {
-            push(@missing_attrs, $attr) unless (exists $options->{$attr});
+            my @bits = split /\./, $attr;
+            my $node = $options;
+
+            push(@missing_attrs, $attr)
+                unless @bits == grep {
+                       ref $node eq "HASH"
+                    && exists $node->{$_}
+                    && ($node = $node->{$_} // {})
+                } @bits;
         }
 
         return { error => 'mandatory attributes for this command missing: '
@@ -630,7 +668,7 @@ Web::API - Web::API - A Simple base module to implement almost every RESTful API
 
 =head1 VERSION
 
-version 1.8
+version 1.9
 
 =head1 SYNOPSIS
 
@@ -872,6 +910,18 @@ default: undef
 
 default: undef
 
+=head2 encoder (custom options encoding subroutine)
+
+Receives options and content-type as the only 2 arguments
+
+default: undef
+
+=head2 decoder (custom response content decoding subroutine)
+
+Receives content and content-type as the only 2 arguments
+
+default: undef
+
 =head2 oauth_post_body (required for all oauth_* auth_types)
 
 default: true
